package ru.dbotthepony.mc.otm.container

import com.google.common.collect.ImmutableList
import com.google.common.collect.ImmutableMap
import com.google.common.collect.ImmutableSet
import it.unimi.dsi.fastutil.ints.IntAVLTreeSet
import it.unimi.dsi.fastutil.ints.IntArrayList
import it.unimi.dsi.fastutil.ints.IntOpenHashSet
import it.unimi.dsi.fastutil.ints.IntSet
import it.unimi.dsi.fastutil.objects.Object2ObjectFunction
import it.unimi.dsi.fastutil.objects.Reference2ObjectOpenHashMap
import net.minecraft.world.Container
import net.minecraft.world.entity.player.Player
import net.minecraft.world.item.Item
import net.minecraft.world.item.ItemStack
import ru.dbotthepony.mc.otm.container.util.containerSlot
import ru.dbotthepony.mc.otm.container.util.iterator
import ru.dbotthepony.mc.otm.core.collect.concatIterators
import ru.dbotthepony.mc.otm.core.collect.filter
import ru.dbotthepony.mc.otm.core.collect.flatMap
import ru.dbotthepony.mc.otm.core.collect.map
import ru.dbotthepony.mc.otm.core.isNotEmpty
import ru.dbotthepony.mc.otm.core.stream
import java.util.stream.Stream

class CombinedContainer<S : IContainerSlot>(containers: Stream<Pair<IEnhancedContainer<S>, Iterable<Int>>>) : ISlottedContainer<S> {
	constructor(vararg containers: IEnhancedContainer<S>) : this(containers.stream().map { it to (0 until it.containerSize) })
	constructor(containers: Collection<IEnhancedContainer<S>>) : this(containers.stream().map { it to (0 until it.containerSize) })

	private val slots: ImmutableList<S>
	private val slotsMap: ImmutableMap<Container, List<S>>
	private val containers: ImmutableSet<Container>
	private val fullCoverage: ImmutableList<Container>
	private val notFullCoverage: ImmutableMap<Container, List<S>>

	init {
		val list = ImmutableList.Builder<S>()
		val validationMap = Reference2ObjectOpenHashMap<Container, IntSet>()
		val slotsMap = Reference2ObjectOpenHashMap<Container, ArrayList<S>>()
		var i = 0

		for ((container, slots) in containers) {
			val validator = validationMap.computeIfAbsent(container, Object2ObjectFunction { IntOpenHashSet() })
			val slotList = slotsMap.computeIfAbsent(container, Object2ObjectFunction { ArrayList() })

			for (slot in slots) {
				if (validator.add(slot)) {
					val slotObj = container.containerSlot(slot)
					list.add(slotObj)
					slotList.add(slotObj)
				} else {
					throw IllegalArgumentException("Duplicate mapping for $container at $i for slot $slot")
				}

				i++
			}
		}

		slots = list.build()
		this.containers = ImmutableSet.copyOf(validationMap.keys)

		this.slotsMap = slotsMap.entries
			.stream()
			.map { it.key to ImmutableList.copyOf(it.value) }
			.collect(ImmutableMap.toImmutableMap({ it.first }, { it.second }))

		this.fullCoverage = this.slotsMap.entries
			.stream()
			.filter { it.value.size == it.key.containerSize }
			.map { it.key }
			.collect(ImmutableList.toImmutableList())

		this.notFullCoverage = this.slotsMap.entries
			.stream()
			.filter { it.key !in this.fullCoverage }
			.collect(ImmutableMap.toImmutableMap({ it.key }, { it.value }))
	}

	override fun clearContent() {
		for (container in fullCoverage) {
			container.clearContent()
		}

		for (slots in notFullCoverage.values) {
			for (slot in slots) {
				slot.remove()
			}
		}
	}

	override fun getContainerSize(): Int {
		return slots.size
	}

	override fun isEmpty(): Boolean {
		return fullCoverage.all { it.isEmpty } && notFullCoverage.values.all { it.all { it.isEmpty } }
	}

	override fun setChanged() {
		for (container in containers) {
			container.setChanged()
		}
	}

	override fun stillValid(player: Player): Boolean {
		return containers.all { it.stillValid(player) }
	}

	override fun iterator(): Iterator<ItemStack> {
		if (notFullCoverage.isEmpty())
			return fullCoverage.iterator().flatMap { it.iterator() }

		return concatIterators(
			fullCoverage.iterator().flatMap { it.iterator() },
			notFullCoverage.values.iterator().flatMap { it.iterator() }.map { it.item }.filter { it.isNotEmpty }
		)
	}

	override fun slotIterator(): Iterator<S> {
		return slots.iterator()
	}

	override fun containerSlot(slot: Int): S {
		return slots[slot]
	}

	override fun setChanged(slot: Int) {
		slots[slot].setChanged()
	}

	class Builder<S : IContainerSlot> {
		private val values = ArrayList<Pair<IEnhancedContainer<S>, Iterable<Int>>>()

		fun add(container: Container): Builder<S> {
			return add(IEnhancedContainer.wrap(container))
		}

		fun add(container: Container, slots: Iterator<Int>): Builder<S> {
			return add(IEnhancedContainer.wrap(container), slots)
		}

		fun add(container: Container, slot: Int): Builder<S> {
			return add(IEnhancedContainer.wrap(container), slot)
		}

		fun add(container: Container, from: Int, to: Int): Builder<S> {
			return add(IEnhancedContainer.wrap(container), from, to)
		}

		fun add(container: Container, slots: Iterable<Int>): Builder<S> {
			return add(IEnhancedContainer.wrap(container), slots)
		}

		fun add(container: IEnhancedContainer<S>): Builder<S> {
			values.add(container to container.slotRange)
			return this
		}

		fun add(container: IEnhancedContainer<S>, slots: Iterator<Int>): Builder<S> {
			values.add(container to IntArrayList(slots))
			return this
		}

		fun add(container: IEnhancedContainer<S>, slot: Int): Builder<S> {
			values.add(container to intArrayOf(slot).asIterable())
			return this
		}

		fun add(container: IEnhancedContainer<S>, from: Int, to: Int): Builder<S> {
			values.add(container to (from .. to))
			return this
		}

		fun add(container: IEnhancedContainer<S>, slots: Iterable<Int>): Builder<S> {
			values.add(container to slots)
			return this
		}

		fun build(): CombinedContainer<S> {
			return CombinedContainer(values.stream())
		}
	}

	companion object {
		fun fromMenuSlots(slots: Iterator<net.minecraft.world.inventory.Slot>): CombinedContainer<IContainerSlot> {
			val builder = Builder<IContainerSlot>()

			for (slot in slots) {
				builder.add(slot.container, slot.slotIndex)
			}

			return builder.build()
		}
	}
}
